# w.c.s. - web application for online forms
# Copyright (C) 2005-2018  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from django.utils.functional import SimpleLazyObject

from quixote import get_publisher

import qommon.misc
from qommon.evalutils import make_datetime
from qommon.templatetags.qommon import parse_datetime


class LazyFormDef(object):
    def __init__(self, formdef):
        self.formdef = formdef

    @property
    def name(self):
        return self.formdef.name

    @property
    def slug(self):
        return self.formdef.url_name

    @property
    def objects(self):
        from wcs.formdef import FormDefSubstVar
        return FormDefSubstVar(self.formdef)

    @property
    def option(self):
        return LazyFormDefOption(self.formdef)


class LazyFormDefOption(object):
    def __init__(self, formdef):
        self.formdef = formdef

    _options = None
    @property
    def options(self):
        if self._options is not None:
            return self._options
        self._options = {}
        if not self.formdef.workflow.variables_formdef:
            return self._options
        if not self.formdef.workflow_options:
            return self._options
        for field in self.formdef.workflow.variables_formdef.fields:
            if not field.varname:
                continue
            self._options[field.varname] = self.formdef.workflow_options.get(field.varname)
            if field.store_display_value:
                if '%s_display' % field.varname in self.formdef.workflow_options:
                    self._options[field.varname + '_raw'] = self._options[field.varname]
                    self._options[field.varname] = self.formdef.workflow_options[
                            '%s_display' % field.varname]
            if field.store_structured_value:
                if '%s_structured' % field.varname in self.formdef.workflow_options:
                    self._options[field.varname + '_structured'] = self.formdef.workflow_options.get(
                            '%s_structured' % field.varname)
        return self._options

    def __getitem__(self, key):
        return self.options[key]


class LazyFormData(LazyFormDef):
    def __init__(self, formdata):
        super(LazyFormData, self).__init__(formdata.formdef)
        self.formdata = formdata

    @property
    def receipt_date(self):
        return qommon.misc.strftime(qommon.misc.date_format(), self.formdata.receipt_time)

    @property
    def receipt_time(self):
        return qommon.misc.strftime('%H:%M', self.formdata.receipt_time)

    @property
    def number(self):
        return self.formdata.get_display_id()

    @property
    def number_raw(self):
        return str(self.formdata.id) if self.formdata.id else None

    @property
    def url(self):
        return self.formdata.get_url()

    @property
    def url_backoffice(self):
        return self.formdata.get_url(backoffice=True)

    @property
    def backoffice_url(self):
        return self.formdata.get_url(backoffice=True)

    @property
    def uri(self):
        return '%s/%s/' % (self.formdef.url_name, self.formdata.id)

    @property
    def criticality_level(self):
        return self.formdata.criticality_level

    @property
    def criticality_label(self):
        try:
            return self.formdata.get_criticality_level_object().name
        except IndexError:
            return None

    @property
    def digest(self):
        return self.formdata.digest

    @property
    def receipt_datetime(self):
        return make_datetime(self.formdata.receipt_time) if self.formdata.receipt_time else None

    @property
    def status(self):
        return self.formdata.get_status_label()

    @property
    def status_is_endpoint(self):
        return self.formdata.is_at_endpoint_status()

    @property
    def tracking_code(self):
        return self.formdata.tracking_code

    @property
    def submission_backoffice(self):
        return self.formdata.backoffice_submission

    @property
    def submission_channel(self):
        return self.formdata.submission_channel

    @property
    def submission_channel_label(self):
        return self.formdata.get_submission_channel_label()

    @property
    def submission_context(self):
        return self.formdata.submission_context

    @property
    def status_url(self):
        return '%sstatus' % self.formdata.get_url()

    @property
    def details(self):
        return self.formdef.get_detailed_email_form(self.formdata, self.formdata.get_url())

    @property
    def user(self):
        user = self.formdata.get_user()
        return LazyUser(user) if user else None

    @property
    def var(self):
        return LazyFormDataVar(self.formdef.get_all_fields(), self.formdata.data, self.formdata)

    @property
    def field(self):
        # no lazy dictionary here as it's legacy.
        d = {}
        for field in self.formdef.get_all_fields():
            if not hasattr(field, 'get_view_value'):
                continue
            value = self.formdata.data.get(field.id)
            if value is not None and field.convert_value_to_str:
                value = field.convert_value_to_str(value)
            elif value is None:
                value = ''
            identifier_name = qommon.misc.simplify(field.label, space='_')
            d[identifier_name] = value

        return d

    @property
    def role(self):
        from wcs.roles import Role
        workflow_roles = {}
        if self.formdef.workflow_roles:
            workflow_roles.update(self.formdef.workflow_roles)
        if self.formdata.workflow_roles:
            workflow_roles.update(self.formdata.workflow_roles)

        d = {}
        for role_type, role_id in workflow_roles.items():
            prefix = '%s_' % role_type.replace('-', '_').strip('_')
            try:
                d.update(Role.get(role_id).get_substitution_variables(prefix))
            except KeyError:
                pass
        return d

    @property
    def comment(self):
        if self.formdata.evolution and self.formdata.evolution[-1].comment:
            return self.formdata.evolution[-1].comment
        return ''

    @property
    def attachments(self):
        from wcs.workflows import AttachmentsSubstitutionProxy
        return AttachmentsSubstitutionProxy(self.formdata)

    @property
    def geoloc(self):
        data = {}
        if self.formdata.geolocations:
            for k, v in self.formdata.geolocations.items():
                data[k] = v
                data[k + '_lat'] = v.get('lat')
                data[k + '_lon'] = v.get('lon')
        return data

    @property
    def previous_status(self):
        if self.formdata.evolution:
            for evolution in reversed(self.formdata.evolution):
                if evolution.status and evolution.status != self.formdata.status:
                    return self.formdata.get_status_label(evolution.status)
        return ''

    @property
    def status_changed(self):
        return self.status != self.previous_status

    @property
    def evolution(self):
        return self.formdef.get_detailed_evolution(self.formdata)

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            if key.startswith('f'):
                for field in self.formdef.get_all_fields():
                    if str(field.id) == str(key[1:]):
                        return self.formdata.data.get(field.id)
            raise


class LazyFormDataVar(object):
    def __init__(self, fields, data, formdata=None):
        self.fields = fields
        self.data = data or {}
        self.formdata = formdata

    _varnames = None
    @property
    def varnames(self):
        if self._varnames is not None:
            return self._varnames
        self._varnames = {}
        for field in self.fields:
            if not field.varname:
                continue
            self._varnames[field.varname] = field
        return self._varnames

    def __getitem__(self, key):
        try:
            field = self.varnames[key]
        except KeyError:
            # key was unknown but for some  values we may still have to provide
            # multiple keys (for example file fields will expect to have both
            # form_var_foo and form_var_foo_raw set to None) and user
            # conditions may use the "is" operator that cannot be overridden
            # (this applies to None as well as boolean values).
            #
            # Therefore we catch unknown keys with known suffixes ("foo_raw")
            # and remove the suffix to get the actual field.  If the data is
            # None or a boolean type, we return it as is.
            if not (key.endswith('_raw') or key.endswith('_url')):
                raise
            maybe_varname = key.rsplit('_', 1)[0]
            field = self.varnames[maybe_varname]
            if self.data.get(field.id) in (None, True, False):
                # valid suffix and data of the correct type
                return self.data.get(field.id)
            raise KeyError(key)

        if self.data.get(field.id) is None:
            return None

        if str(field.id) not in self.data:
            raise KeyError(key)

        if field.key == 'date':
            return LazyFieldVarDate(self.data, field, self.formdata)
        if field.key == 'map':
            return LazyFieldVarMap(self.data, field, self.formdata)

        return LazyFieldVar(self.data, field, self.formdata)

    def __getattr__(self, attr):
        try:
            return self.__getitem__(attr)
        except KeyError:
            raise AttributeError(attr)


class LazyFieldVar(object):
    def __init__(self, data, field, formdata=None):
        self.data = data
        self.field = field
        self.formdata = formdata

    @property
    def raw(self):
        if self.field.store_display_value or self.field.key in ('file', 'date'):
            return self.data.get(self.field.id)
        raise KeyError('raw')

    @property
    def url(self):
        if self.field.key != 'file' or not self.formdata:
            raise KeyError('url')
        return '%sdownload?f=%s' % (self.formdata.get_url(), self.field.id)

    def get_value(self):
        if self.field.store_display_value:
            return self.data.get('%s_display' % self.field.id)
        value = self.data.get(self.field.id)
        if self.field.convert_value_to_str:
            return self.field.convert_value_to_str(value)
        if isinstance(value, str):
            return unicode(value, get_publisher().site_charset)
        return value

    def __str__(self):
        value = self.get_value()
        if isinstance(value, unicode):
            return value.encode(get_publisher().site_charset)
        return str(value)

    def __nonzero__(self):
        if self.field.key == 'bool':
            return bool(self.data.get(self.field.id))
        return bool(self.get_value())

    def __contains__(self, value):
        if self.field.key == 'items':
            return qommon.misc.site_encode(value) in self.data.get(self.field.id)
        return value in self.get_value()

    def __unicode__(self):
        return unicode(str(self), get_publisher().site_charset)

    def __eq__(self, other):
        return unicode(self) == unicode(other)

    def __ne__(self, other):
        return unicode(self) != unicode(other)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.get_value()[key]
        try:
            return getattr(self, key)
        except AttributeError:
            pass
        structured_value = self.field.get_structured_value(self.data)
        if not structured_value:
            raise KeyError(key)
        if isinstance(structured_value, dict):
            return structured_value[key]
        if isinstance(structured_value, list):
            for i, struct_value in enumerate(structured_value):
                if str(key) == str(i):
                    return struct_value
        raise KeyError(key)

    def __iter__(self):
        return iter(self.get_value())

    def __add__(self, other):
        return self.get_value().__add__(other)

    def __mul__(self, other):
        return self.get_value().__mul__(other)

    def __getstate__(self):
        raise AssertionError('lazy cannot be pickled')


class DateOperatorsMixin(object):
    def __eq__(self, other):
        if hasattr(other, 'timetuple'):
            other = other.timetuple()
        elif hasattr(other, 'get_value'):
            other = other.get_value()
        return parse_datetime(self.timetuple()) == parse_datetime(other)

    def __ne__(self, other):
        if hasattr(other, 'timetuple'):
            other = other.timetuple()
        elif hasattr(other, 'get_value'):
            other = other.get_value()
        return parse_datetime(self.timetuple()) != parse_datetime(other)

    def __gt__(self, other):
        if hasattr(other, 'timetuple'):
            other = other.timetuple()
        elif hasattr(other, 'get_value'):
            other = other.get_value()
        return parse_datetime(self.timetuple()) > parse_datetime(other)

    def __lt__(self, other):
        if hasattr(other, 'timetuple'):
            other = other.timetuple()
        elif hasattr(other, 'get_value'):
            other = other.get_value()
        return parse_datetime(self.timetuple()) < parse_datetime(other)

    def __ge__(self, other):
        if hasattr(other, 'timetuple'):
            other = other.timetuple()
        elif hasattr(other, 'get_value'):
            other = other.get_value()
        return parse_datetime(self.timetuple()) >= parse_datetime(other)

    def __le__(self, other):
        if hasattr(other, 'timetuple'):
            other = other.timetuple()
        elif hasattr(other, 'get_value'):
            other = other.get_value()
        return parse_datetime(self.timetuple()) <= parse_datetime(other)


class LazyDateObject(DateOperatorsMixin, SimpleLazyObject):
    pass


def lazy_date(value):
    return LazyDateObject(lambda: value)


class LazyFieldVarDate(DateOperatorsMixin, LazyFieldVar):
    def get_raw(self):
        return self.data.get(self.field.id)

    def timetuple(self):
        return self.get_raw()

    # for backward compatibility with sites using time.struct_time
    # methods we still have to provide time.struct_time properties.
    @property
    def tm_year(self):
        return self.get_raw().tm_year

    @property
    def tm_mon(self):
        return self.get_raw().tm_mon

    @property
    def tm_mday(self):
        return self.get_raw().tm_mday

    @property
    def tm_hour(self):
        return self.get_raw().tm_hour

    @property
    def tm_min(self):
        return self.get_raw().tm_min

    @property
    def tm_sec(self):
        return self.get_raw().tm_sec

    @property
    def tm_wday(self):
        return self.get_raw().tm_wday

    @property
    def tm_yday(self):
        return self.get_raw().tm_yday


class LazyFieldVarMap(LazyFieldVar):
    def split(self, *args, **kwargs):
        # Compatibility with usage of map variable as a string. It is
        # recommended to use lat/lon properties instead.
        return self.data.get(self.field.id).split(*args, **kwargs)


class LazyUser(object):
    def __init__(self, user):
        self.user = user

    @property
    def display_name(self):
        return self.user.display_name

    @property
    def email(self):
        return self.user.email

    @property
    def var(self):
        return LazyFormDataVar(self.user.formdef.fields, self.user.form_data)

    @property
    def admin_access(self):
        return self.user.can_go_in_admin()

    @property
    def backoffice_access(self):
        return self.user.can_go_in_backoffice()

    @property
    def name_identifier(self):
        d = {}
        for i, name_identifier in enumerate(self.user.name_identifiers):
            d[str(i)] = name_identifier
        return d

    def __getitem__(self, key):
        return getattr(self, key)

    def __getattr__(self, attr):
        try:
            return super(LazyUser, self).__getattr__(attr)
        except AttributeError:
            return getattr(self.user, attr)


class LazyRequest(object):
    def __init__(self, request):
        self.request = request

    @property
    def quixote_request(self):  # compatibility
        return self.request

    @property
    def GET(self):
        return self.request.django_request.GET

    @property
    def is_in_backoffice(self):
        return self.request.is_in_backoffice()

    @property
    def method(self):
        return self.request.method

    @property
    def user(self):
        return self.request.user
